local map={
	"1111111111111111",
	"1   12         1",
	"1 111111111111 1",
	"1 1      1   1 1",
	"1 1 1111 1 1 1 1",
	"1 1   1  1 1   1",
	"1 111 111111 111",
	"1   1          1",
	"111 11111111 1 1",
	"1 1          1 1",
	"1 111111111111 1",
	"1   1 1    1   1",
	"1 111 1 1111 111",
	"1     1    1   1",
	"1 11111111 111 1",
	"1              1",
	"1111114  4111111",
	"1              1",
	"1              1",
	"1  3  3  3  3  1",
	"1              1",
	"1              1",
	"1  2  5  5  2  1",
	"1              1",
	"1              1",
	"1  2  5  5  2  1",
	"1              1",
	"1              1",
	"1  3  2  2  3  1",
	"1              1",
	"1              1",
	"1111111111111111",
}

love.mouse.setVisible(false)

local mapW, mapH = #map[1], #map

local realMap = {}

for y=1,mapH do
	local rc = {}
	local r = map[y]
	
	for x=1,mapW do
		rc[x] = tonumber(r:sub(x,x))
	end
	
	realMap[y] = rc
end

local colorSchemes = {
	{0,8}, -- White+gray
	{3,11}, -- Blue
	{6,14}, -- Red
	{5,13}, -- Green
	{4,1}, -- Yellow/orange
}

local cos, sin, floor = math.cos, math.sin, math.floor

local function cast(cx,cy,angle)
	-- Direction vector
	local vx,vy=cos(angle), sin(angle)
	local slope=vy/vx
	
	-- Next distance, x and y axis points
	local ndx, ndy
	
	-- Steps, distance and block
	local dsx, dsy, bsx, bsy
	
	if vx < 0 then
		local x = ( cx % 1 )
		bsx = -1
		ndx = ( x * x * ( 1 + slope * slope ) ) ^ 0.5
		dsx = ( ( 1 + slope * slope ) ) ^ 0.5
	else
		local x=1-(cx%1)
		bsx=1
		ndx=(x*x*(1+slope*slope))^0.5
		dsx=((1+slope*slope))^0.5
	end

	if vy<0 then
		local y=(cy%1)
		bsy=-1
		ndy=(y*y*(1+1/(slope*slope)))^0.5
		dsy=((1+1/(slope*slope)))^0.5
	else
		local y=1-(cy%1)
		bsy=1
		ndy=(y*y*(1+1/(slope*slope)))^0.5
		dsy=((1+1/(slope*slope)))^0.5
	end

	local x,y=floor(cx),floor(cy)
	local cury = realMap[y]
	while x>0 and x<=mapW and y>0 and y<=mapH do
		local hitD
		local isX
		if ndx<ndy then
			-- X crossing is next
			x=x+bsx
			isX=true
			hitD=ndx
			ndx=ndx+dsx
		else
			y=y+bsy
			cury = realMap[y]
			isX=false
			hitD=ndy
			ndy=ndy+dsy
		end
		local wall=cury[x]
		if wall then
			return colorSchemes[wall][isX and 1 or 2], hitD
		end
	end  
end

local w,h=love.window.getDimensions( )
local centerX, centerY=math.floor((w+1)/2), math.floor((h+1)/2)

local px, py=9,25
local dir=0
local fx,fy
local speed=.1
local turnSpeed=4
local mouseSpeed=0.5

local function turn(amt)
	dir=dir+amt
	fx,fy=math.cos(math.rad(dir)), math.sin(math.rad(dir))
end

turn(0)

-- Build table of angles and base distances per scanline
local screenDist = 0.55 * w
local scan = {}

for x = 1, w do
	local t={}
	scan[x] = t
	
	t.angle = math.atan2(x - centerX, screenDist)
	t.dist = ((x - centerX)^2 + screenDist^2)^.5 / screenDist
end

-- Borrowed from CCLite
local COLOUR_CODE = {
	[1] = {240, 240, 240},
	[2] = {242, 178, 51},
	[4] = {229, 127, 216},
	[8] = {153, 178, 242},
	[16] = {222, 222, 108},
	[32] = {127, 204, 25},
	[64] = {242, 178, 204},
	[128] = {76, 76, 76},
	[256] = {153, 153, 153},
	[512] = {76, 153, 178},
	[1024] = {178, 102, 229},
	[2048] = {37, 49, 146},
	[4096] = {127, 102, 76},
	[8192] = {87, 166, 78},
	[16384] = {204, 76, 76},
	[32768] = {0, 0, 0},
}

local function redraw()
	for x = 1, w do
		local wall, dist = cast(px, py, math.rad(dir) + scan[x].angle)
		if wall then
			-- Calc wall height based on distance
			local height = scan[x].dist / dist
			height = math.floor(math.min(height * centerY, (h + 1) / 2))
			love.graphics.setColor(COLOUR_CODE[2 ^ wall])
			love.graphics.line(x - 0.5, centerY - height - 0.5, x - 0.5, centerY + height + 0.5)
		end
	end
end

local function clampCollision(x,y,radius)
	-- Am I *in* a block?
	local gx,gy=math.floor(x),math.floor(y)
	if map[gy]:sub(gx,gx)~=" " then
		-- I am. Complete fail, do nothing.
		return x,y
	end
	
	-- ok, check the neighbors.
	local right=math.floor(x+radius)>gx
	local left=math.floor(x-radius)<gx
	local front=math.floor(y-radius)<gy
	local back=math.floor(y+radius)>gy
	
	local pushed=false
	
	if right and map[gy]:sub(gx+1,gx+1)~=" " then
		-- push left
		pushed=true
		x=gx+1-radius
	elseif left  and map[gy]:sub(gx-1,gx-1)~=" " then
		-- push right
		pushed=true
		x=gx+radius
	end
	
	if front and map[gy-1]:sub(gx,gx)~=" " then
		-- push back
		pushed=true
		y=gy+radius
	elseif back and map[gy+1]:sub(gx,gx)~=" " then
		-- push forward
		pushed=true
		y=gy+1-radius
	end
 
	-- if I wasn't pushed out on any side, I might be hitting a corner
	if not pushed then
		-- square rad
		local r2=radius^2
		local pushx,pushy=0,0
		if left then
			if front and map[gy-1]:sub(gx-1,gx-1)~=" " then
				-- check front-left
				local dist2=(gx-x)^2+(gy-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=pushd,pushd
				end
			elseif back and map[gy+1]:sub(gx-1,gx-1)~=" " then
				local dist2=(gx-x)^2+(gy+1-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=pushd,-pushd
				end
			end
		elseif right then
			if front and map[gy-1]:sub(gx+1,gx+1)~=" " then
				-- check front-left
				local dist2=(gx+1-x)^2+(gy-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=-pushd,pushd
				end
			elseif back and map[gy+1]:sub(gx+1,gx+1)~=" " then
				local dist2=(gx+1-x)^2+(gy+1-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=-pushd,-pushd
				end
			end
		end
		x=x+pushx
		y=y+pushy
	end
	
	return x,y
end

local dirty=true

local prevX, prevY = w/2, h/2

function love.update(dt)
	local modifier = math.min(dt / 0.05,1)
	
	if love.keyboard.isDown("left") then
		turn(-turnSpeed * modifier)
		dirty = true
	end
	
	if love.keyboard.isDown("right") then
		turn(turnSpeed * modifier)
		dirty = true
	end
	
	if love.keyboard.isDown("up") or love.keyboard.isDown("w") then
		px = px + fx * speed * modifier
		py = py + fy * speed * modifier
		dirty = true
	end
	
	if love.keyboard.isDown("down") or love.keyboard.isDown("s") then
		px = px - fx * speed * modifier
		py = py - fy * speed * modifier
		dirty = true
	end
	
	if love.keyboard.isDown("a") then
		px = px + fy * speed * modifier
		py = py - fx * speed * modifier
		dirty = true
	end
	
	if love.keyboard.isDown("d") then
		px = px - fy * speed * modifier
		py = py + fx * speed * modifier
		dirty = true
	end
	
	if love.keyboard.isDown("q") or love.keyboard.isDown("escape") then
		love.event.quit()
	end
	
	local mX, mY = love.mouse.getPosition()
	if mX ~= prevX or mY ~= prevY then
		if mX < prevX then
			turn(-(mouseSpeed * (prevX - mX)) * modifier)
			dirty = true
		end
		
		if mX > prevX then
			turn((mouseSpeed * (mX - prevX)) * modifier)
			dirty = true
		end
	end
	
	love.mouse.setPosition( w/2, h/2 )
end

love.graphics.setLineStyle("rough")
love.graphics.setBackgroundColor(76, 76, 76)

local present = love.graphics.present
local clear = love.graphics.clear
love.graphics.clear = function() end
love.graphics.present = function() end

function love.draw()
	px, py = clampCollision(px, py, .25)
	if dirty then
		clear()
		redraw()
		present()
		dirty = false
	end
end